@@ -1,5 +1,6 @@
 use super::{Context, Options};
 use crate::command::release_impl::tag_name_for;
+use crate::command::release_impl::utils::will;
 use anyhow::{anyhow, bail};
 use bstr::ByteSlice;
 use cargo_metadata::{
@@ -12,7 +13,7 @@ use git_repository::{
     object,
     odb::{pack, Find, FindExt},
     refs::{
-        file,
+        self, file,
         file::loose::reference::peel,
         mutable::Target,
         transaction::{Change, Create, RefEdit},
@@ -24,15 +25,18 @@ use std::{convert::TryInto, process::Command};
 pub(in crate::command::release_impl) fn has_changed_since_last_release(
     package: &Package,
     ctx: &Context,
+    verbose: bool,
 ) -> anyhow::Result<bool> {
     let version_tag_name = tag_name_for(&package.name, &package.version.to_string());
     let mut tag_ref = match ctx.repo.refs.find(&version_tag_name, ctx.packed_refs.as_ref())? {
         None => {
-            log::info!(
-                "Package {} wasn't tagged with {} yet and thus needs a release",
-                package.name,
-                version_tag_name
-            );
+            if verbose {
+                log::info!(
+                    "Package {} wasn't tagged with {} yet and thus needs a release",
+                    package.name,
+                    version_tag_name
+                );
+            }
             return Ok(true);
         }
         Some(r) => r,
@@ -47,22 +51,26 @@ pub(in crate::command::release_impl) fn has_changed_since_last_release(
     let target = peel_ref_fully(&mut ctx.repo.refs.find_existing("HEAD", None)?, ctx)?;
     let released_target = peel_ref_fully(&mut tag_ref, ctx)?;
 
-    let mut buf = Vec::new();
+    if repo_relative_crate_dir.as_os_str().is_empty() {
+        Ok(target != released_target)
+    } else {
+        let mut buf = Vec::new();
 
-    let current_dir_id = find_directory_id_in_tree(
-        repo_relative_crate_dir,
-        resolve_tree_id_from_ref_target(target, &ctx.repo, &mut buf)?,
-        &ctx.repo,
-        &mut buf,
-    )?;
-    let released_dir_id = find_directory_id_in_tree(
-        repo_relative_crate_dir,
-        resolve_tree_id_from_ref_target(released_target, &ctx.repo, &mut buf)?,
-        &ctx.repo,
-        &mut buf,
-    )?;
+        let current_dir_id = find_directory_id_in_tree(
+            repo_relative_crate_dir,
+            resolve_tree_id_from_ref_target(target, &ctx.repo, &mut buf)?,
+            &ctx.repo,
+            &mut buf,
+        )?;
+        let released_dir_id = find_directory_id_in_tree(
+            repo_relative_crate_dir,
+            resolve_tree_id_from_ref_target(released_target, &ctx.repo, &mut buf)?,
+            &ctx.repo,
+            &mut buf,
+        )?;
 
-    Ok(released_dir_id != current_dir_id)
+        Ok(released_dir_id != current_dir_id)
+    }
 }
 
 pub fn assure_clean_working_tree() -> anyhow::Result<()> {
@@ -188,34 +196,65 @@ pub fn create_version_tag(
     new_version: &str,
     commit_id: ObjectId,
     repo: &Repository,
-    Options { dry_run, skip_tag, .. }: Options,
-) -> anyhow::Result<()> {
+    Options {
+        verbose,
+        dry_run,
+        skip_tag,
+        ..
+    }: Options,
+) -> anyhow::Result<Option<refs::mutable::FullName>> {
     if skip_tag {
-        return Ok(());
+        return Ok(None);
     }
     let tag_name = tag_name_for(&publishee.name, new_version);
+    let edit = RefEdit {
+        change: Change::Update {
+            log: Default::default(),
+            mode: Create::Only,
+            new: Target::Peeled(commit_id),
+        },
+        name: format!("refs/tags/{}", tag_name).try_into()?,
+        deref: false,
+    };
     if dry_run {
-        log::info!("WOULD create tag {}", tag_name);
+        if verbose {
+            log::info!("WOULD create tag {}", tag_name);
+        }
+        Ok(Some(edit.name))
     } else {
-        for tag in repo
+        let edits = repo
             .refs
             .transaction()
-            .prepare(
-                Some(RefEdit {
-                    change: Change::Update {
-                        log: Default::default(),
-                        mode: Create::Only,
-                        new: Target::Peeled(commit_id),
-                    },
-                    name: format!("refs/tags/{}", tag_name).try_into()?,
-                    deref: false,
-                }),
-                git_lock::acquire::Fail::Immediately,
-            )?
-            .commit(&actor::Signature::empty())?
-        {
-            log::info!("Created tag {}", tag.name.as_bstr());
-        }
+            .prepare(Some(edit), git_lock::acquire::Fail::Immediately)?
+            .commit(&actor::Signature::empty())?;
+        assert_eq!(edits.len(), 1, "We create only one tag and there is no expansion");
+        let tag = edits.into_iter().next().expect("the promised tag");
+        log::info!("Created tag {}", tag.name.as_bstr());
+        Ok(Some(tag.name))
+    }
+}
+
+// TODO: Make this gitoxide
+pub fn push_tags_and_head(
+    tag_names: impl IntoIterator<Item = refs::mutable::FullName>,
+    options: Options,
+) -> anyhow::Result<()> {
+    if options.skip_push {
+        return Ok(());
+    }
+
+    let mut cmd = Command::new("git");
+    cmd.arg("push").arg("HEAD");
+    for tag_name in tag_names {
+        cmd.arg(tag_name.as_bstr().to_str()?);
+    }
+
+    if options.verbose {
+        log::info!("{} run {:?}", will(options.dry_run), cmd);
+    }
+    if options.dry_run || cmd.status()?.success() {
+        Ok(())
+    } else {
+        bail!("'git push' invocation failed. Try to push manually and repeat the smart-release invocation to resume.");
     }
-    Ok(())
 }
